﻿using System;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Threading;

namespace UTScreenshotFixer
{
    internal static class Program
    {
        private struct ConstantHeaderPart
        {
            public readonly int Offset;
            public readonly byte[] Values;

            public ConstantHeaderPart(int offset, byte[] values)
            {
                Offset = offset;
                Values = values;
            }
        }

        private static readonly ConstantHeaderPart[] ConstantHeaderParts =
        {
            new ConstantHeaderPart(0, new byte[] {0x42, 0x4D}),
            new ConstantHeaderPart(6, new byte[] {0, 0, 0, 0, 54, 0, 0, 0, 40, 0, 0, 0}),
            new ConstantHeaderPart(26, new byte[] {1, 0, 24, 0, 0, 0, 0, 0}),
            new ConstantHeaderPart(38, new byte[] {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}),
        };

        private static int numTotal;
        private static int numFixed;

        private static void Main(string[] args)
        {
            if (args.Length == 0)
                Console.WriteLine("No files specified.");

            numTotal = 0;
            numFixed = 0;
            var timer = Stopwatch.StartNew();
            foreach(var fileName in args)
                ProcessFile(fileName);

            Console.WriteLine("{0} files processed in {2}, {1} were fixed.", numTotal, numFixed, timer.Elapsed);
        }

        private static void ProcessFile(string fileName)
        {
            if (!File.Exists(fileName))
            {
                Console.WriteLine("{0} does not exist!", fileName);
            }
            else
            {
                var data = File.ReadAllBytes(fileName);
                Interlocked.Increment(ref numTotal);

                int width, height;
                if (!IsValidHeader(data, out width, out height))
                {
                    Console.WriteLine("{0} - fix not applicable", fileName);
                    return;
                }

                var rowLength = 3*width;
                var targetRowLength = ((3*width+3)/4)*4;

                if (rowLength == targetRowLength)
                {
                    Console.WriteLine("{0} - no need for fixing", fileName);
                    return;
                }

                var targetDataSize = targetRowLength*height;
                var targetFileSize = 54 + targetDataSize;
                var targetData = new byte[targetFileSize];

                // copy header
                Array.Copy(data, targetData, 54);

                // adjust lengths in header
                Array.Copy(BitConverter.GetBytes(targetFileSize), 0, targetData, 2, 4);
                Array.Copy(BitConverter.GetBytes(targetDataSize), 0, targetData, 34, 4);

                // copy rows, leaving the padding bytes zero
                for (var row = 0; row < height; row++)
                    Array.Copy(data, 54 + row*rowLength, targetData, 54 + row*targetRowLength, rowLength);

                var fixedFileName = Path.ChangeExtension(fileName, ".fixed.bmp");
                File.WriteAllBytes(fixedFileName, targetData);
                Console.WriteLine("{0} -> {1}", fileName, Path.GetFileName(fixedFileName));
                Interlocked.Increment(ref numFixed);
            }
        }

        private static bool IsValidHeader(byte[] data, out int width, out int height)
        {
            if (data.Length < 54)
            {
                width = 0;
                height = 0;
                return false;
            }

            width = BitConverter.ToInt32(data, 18);
            height = BitConverter.ToInt32(data, 22);

            var headerFileSize = BitConverter.ToInt32(data, 2);
            var headerDataSize = BitConverter.ToInt32(data, 34);

            return width > 0 && height > 0 && data.Length == 54 + 3*width*height && headerFileSize == data.Length &&
                   headerDataSize == 3*width*height && ConstantHeaderPartsMatch(data);
        }

        private static bool ConstantHeaderPartsMatch(byte[] data)
        {
            return ConstantHeaderParts.All(
                constantHeaderPart =>
                    constantHeaderPart.Values.SequenceEqual(
                        data.Skip(constantHeaderPart.Offset).Take(constantHeaderPart.Values.Length)));
        }
    }
}
